import abc

from trypnv.machine import may_handle, message  # type: ignore
from trypnv import NvimFacade

from tryp import F, List, Map, _

from proteome.state import ProteomeComponent, ProteomeTransitions
from proteome.logging import Logging

def unite_msg(name):
    return message(name, varargs='unite_args')

UniteSelectAdd = unite_msg('UniteSelectAdd')
UniteSelectAddAll = unite_msg('UniteSelectAddAll')
UniteProjects = unite_msg('UniteProjects')


class UniteEntity(Logging, metaclass=abc.ABCMeta):
    ''' set up sources and kinds dynamically.
    The python call API cannot be used, as funcrefs cannot be
    serialized.
    The callback functions must be called once so that exists()
    can see them, otherwise Unite refuses to work.
    They must be called a/sync according to their definition,
    otherwise it will silently deadlock!
    '''

    def __init__(self, name: str) -> None:
        self.name = name

    @abc.abstractproperty
    def tpe(self):
        ...

    @abc.abstractproperty
    def data(self):
        ...

    @property
    def _func_defs_sync(self):
        return List()

    @property
    def _func_defs_async(self):
        return List()

    def _force_function_defs(self, vim):
        force = lambda c, n: c('silent call {}()'.format(n))
        self._func_defs_sync.foreach(F(force, vim.cmd_sync))
        self._func_defs_async.foreach(F(force, vim.cmd))

    def define(self, vim: NvimFacade):
        self._force_function_defs(vim)
        vim.cmd('call unite#define_{}({})'.format(self.tpe, self.data))


class UniteSource(UniteEntity):
    _templ = '''
    {{
        'name': '{}',
        'gather_candidates': function('{}'),
        'default_kind': '{}',
    }}
    '''.replace('\n', '')

    def __init__(self, name: str, source: str, kind: str) -> None:
        super().__init__(name)
        self.source = source
        self.kind = kind

    @property
    def tpe(self):
        return 'source'

    @property
    def _func_defs_sync(self):
        return List(self.source)

    @property
    def data(self):
        return self._templ.format(self.name, self.source, self.kind)


class UniteKind(UniteEntity):
    _templ = '''
    {{
        'name': '{name}',
        'default_action': '{default}',
        'parents': [],
        'action_table': {{
            {actions}
        }}
    }}
    '''.replace('\n', '')

    _action_templ = '''
    '{name}': {{
        'func': function('{handler}'),
        'description': '{desc}',
    }}
    '''.replace('\n', '')

    @property
    def tpe(self):
        return 'kind'

    def __init__(self, name: str, actions: List[Map]) -> None:
        super().__init__(name)
        self.actions = actions
        self.default = actions.head / _['name'] | 'none'

    @property
    def _func_defs_async(self):
        return self.actions / _['handler']

    def _action(self, params):
        return self._action_templ.format(**params)

    @property
    def data(self):
        actions = self.actions.map(self._action).join(', ')
        return self._templ.format(name=self.name, actions=actions,
                                  default=self.default)


class Plugin(ProteomeComponent):
    addable_candidates = '_proteome_unite_addable'
    all_addable_candidates = '_proteome_unite_all_addable'
    projects_candidates = '_proteome_unite_projects'

    add_project = '_proteome_unite_add_project'
    remove_project = '_proteome_unite_remove_project'
    activate_project = '_proteome_unite_activate_project'

    addable = 'proteome_addable'
    all_addable = 'proteome_all_addable'
    projects = 'proteome_projects'
    project = 'proteome_project'

    unite_msgs = [UniteSelectAdd, UniteSelectAddAll, UniteProjects]

    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)
        self._unite_ready = False

    def prepare(self, msg):
        if (not self._unite_ready and
                type(msg) in self.unite_msgs):
            self._setup_unite()

    def _setup_unite(self):
        addable = UniteSource(self.addable, self.addable_candidates,
                              self.addable)
        all_addable = UniteSource(self.all_addable,
                                  self.all_addable_candidates, self.addable)
        add_action = Map(name='add', handler=self.add_project,
                         desc='add project')
        add_pro = UniteKind(self.addable, List(add_action))
        projects = UniteSource(self.projects, self.projects_candidates,
                               self.project)
        remove_action = Map(name='remove', handler=self.remove_project,
                            desc='remove project')
        activate_action = Map(name='activate', handler=self.activate_project,
                              desc='activate project')
        project = UniteKind(self.project, List(activate_action, remove_action))
        addable.define(self.vim)
        all_addable.define(self.vim)
        add_pro.define(self.vim)
        projects.define(self.vim)
        project.define(self.vim)
        self._unite_ready = True

    class Transitions(ProteomeTransitions):

        def unite_cmd(self, cmd):
            args = ' '.join(self.msg.unite_args)
            self.vim.cmd('Unite {} {}'.format(cmd, args))

        @may_handle(UniteSelectAdd)
        def select_add(self):
            self.unite_cmd(self.machine.addable)

        @may_handle(UniteSelectAddAll)
        def select_add_all(self):
            self.unite_cmd(self.machine.all_addable)

        @may_handle(UniteProjects)
        def projects(self):
            self.unite_cmd(self.machine.projects)

__all__ = ('Plugin', 'UniteSelectAdd', 'UniteSelectAddAll', 'UniteProjects')



from pathlib import Path
from datetime import datetime

from fn import _, F  # type: ignore

from tryp.lazy import lazy
from tryp import Map, __, Just, Empty, may, List, Maybe
from tryp.util.numeric import try_convert_int

from trypnv.machine import (may_handle, message, handle, IO)  # type: ignore
from trypnv.machine import Error  # type: ignore
from trypnv.record import field, dfield, Record, lazy_list_field
from trypnv.nvim import ScratchBuilder, ScratchBuffer

from proteome.state import ProteomeComponent, ProteomeTransitions
from proteome.git import (History, HistoryT, HistoryState, Repo, CommitInfo,
                          HistoryGit)
from proteome.plugins.core import Save, StageIV
from proteome.logging import Logging
from proteome.project import Project
from proteome.plugins.history.messages import (HistoryPrev, HistoryNext,
                                               HistoryStatus, HistoryLog,
                                               HistoryBrowse,
                                               HistoryBrowseInput,
                                               HistorySwitch, Redraw,
                                               QuitBrowse, Commit,
                                               HistoryBufferNext,
                                               HistoryBufferPrev, HistoryPick,
                                               HistoryRevert)
from proteome.plugins.history.patch import Patch


class BrowseState(Record):
    repo = field(Repo)
    current = field(int)
    commits = lazy_list_field()
    buffer = field(ScratchBuffer)
    selected = dfield(0)

Init = message('Init')


class BrowseMachine(ProteomeComponent):
    _data_type = BrowseState

    class Transitions(ProteomeTransitions):

        @property
        def buffer(self):
            return self.data.buffer.proxy

        @lazy
        def content(self):
            sel = self.data.selected
            return List.wrap(enumerate(self.data.commits[:sel + 20]))\
                .flat_smap(lambda i, a: a.browse_format(i == sel))

        def _create_mappings(self):
            List.wrap('jksprq').foreach(self._create_mapping)
            self._create_mapping('<cr>', to='%CR%')

        def _create_mapping(self, keyseq, mode='n', to=None):
            to_seq = Maybe(to) | keyseq
            raw = self.buffer.buffer
            cmd = ':ProHistoryBrowseInput {}<cr>'.format(to_seq)
            raw.nmap(keyseq, cmd)

        def _configure_appearance(self):
            self.buffer.set_options('filetype', 'diff')
            self.buffer.set_options('syntax', 'diff')
            self.buffer.window.set_optionb('cursorline', True)
            self.vim.doautocmd('FileType')
            sy = self.buffer.syntax
            id_fmt = '[a-f0-9]\+'
            sy.match('Commit', '^[* ]\s\+{}'.format(id_fmt),
                     contains='Star,Sha')
            sy.match('Star', '^*', 'contained')
            sy.match('Sha', id_fmt, 'contained')
            sy.link('Star', 'Title')
            sy.link('Sha', 'Identifier')

        @may_handle(Init)
        def browse_init(self):
            self._create_mappings()
            self._configure_appearance()
            return Redraw()

        @may_handle(Redraw)
        def redraw(self):
            self.buffer.set_content(self.content)
            self.vim.cursor(self.data.selected + 1, 1)
            self.vim.feedkeys('zz')

        @handle(HistoryBrowseInput)
        def input(self):
            handlers = Map({
                'j': self._down,
                'k': self._up,
                '%CR%': self._switch,
                's': self._switch,
                'p': self._pick,
                'r': self._revert,
                'q': self._close_tab,
            })
            return handlers.get(self.msg.keyseq).flat_map(lambda f: f())

        def _down(self):
            return self._select_diff(1)

        def _up(self):
            return self._select_diff(-1)

        @may
        def _select_diff(self, diff):
            index = self.data.selected + diff
            if index >= 0 and self.data.commits.min_length(index + 1):
                return self.data.set(selected=index), Redraw()

        @may
        def _switch(self):
            q = QuitBrowse(self.buffer)
            return (HistorySwitch(self.data.selected).pub, q, q.pub)

        @may_handle(QuitBrowse)
        def quit(self):
            if self.msg.buffer == self.buffer:
                self._close_tab()

        @may
        def _close_tab(self):
            self.buffer.tab.close()

        def _pick(self):
            return self._pick_commit(HistoryPick)

        def _revert(self):
            return self._pick_commit(HistoryRevert)

        @may
        def _pick_commit(self, tpe):
            q = QuitBrowse(self.buffer)
            return (tpe(self.data.selected).pub, q, q.pub)


class Browse(Logging):

    def __init__(self, state: BrowseState, vim) -> None:
        self.state = state
        self.machine = BrowseMachine('history_browse', vim)

    @property
    def buffer(self):
        return self.state.buffer

    @property
    def repo(self):
        return self.state.repo

    def run(self):
        return self.send(Init())

    def send(self, msg):
        result = self.machine.loop_process(self.state, msg)
        self.state = result.data
        return result.pub


class Plugin(ProteomeComponent):
    failed_pick_err = 'couldn\'t revert commit'

    @lazy
    def base(self):
        return self.vim.pdir('history_base').get_or_else(Path('/dev/null'))

    class Transitions(ProteomeTransitions):

        def _with_sub(self, state):
            return self.data.with_sub_state(self.name, state)

        @property
        def _with_browse(self):
            return F(self._with_sub) << self.state.setter('browse')

        @property
        def state(self):
            return self.data.sub_state(self.name, HistoryState)

        @lazy
        def history(self):
            return History(self.machine.base, state=self.state)

        @lazy
        def history_t(self):
            return HistoryT(self.history)

        @property
        def all_projects_history(self):
            return self.pflags.all_projects_history

        @lazy
        def projects(self):
            return self.data.history_projects(self.all_projects_history)

        @property
        def current(self):
            return self.data.current

        def _all_projects(self, f):
            new_state = self.projects.fold_left(self.history_t)(f).state
            return self._with_sub(new_state)

        def _with_repos(self, f):
            g = lambda hist, pro: hist / __.at(pro, _ / f)
            new_state = self.projects.fold_left(self.history_t)(g).state
            return self._with_sub(new_state)

        def _with_repo(self, pro, f):
            new_state = (self.history_t / __.at(pro, f)).state
            return self._with_sub(new_state)

        def _with_current_repo(self, f):
            return self.current.map(lambda a: self._with_repo(a, f))

        def _repo_ro(self, project: Project):
            return self.history.repo(project)

        @property
        def _current_repo_ro(self):
            return self.current // self._repo_ro

        @may_handle(StageIV)
        def stage_4(self):
            excludes = self.vim.ppath('history_excludes')
            return self._with_repos(lambda a: a.init(excludes))

        # TODO handle broken repo
        # TODO only save if changes exist
        # TODO allow specifying target
        @may_handle(Commit)
        def commit(self):
            return self._with_repos(__.add_commit_all(self._timestamp))

        def _switch(self, f):
            def notify(repo):
                self.vim.reload_windows()
                repo.current_commit_info\
                    .map(_.num_since)\
                    .foreach(self.log.info)
                return repo
            return self._with_current_repo(_ / f % notify)

        # FIXME save first?
        @handle(HistoryPrev)
        def prev(self):
            return self._switch(__.prev())

        @handle(HistoryNext)
        def next(self):
            return self._switch(__.next())

        @handle(HistorySwitch)
        def switch(self):
            return try_convert_int(self.msg.index)\
                .map(__.select)\
                .flat_map(self._switch)

        @may_handle(HistoryBufferPrev)
        def history_buffer_prev(self):
            pass

        @may_handle(HistoryStatus)
        def history_status(self):
            def log_status(repo):
                if repo.status:
                    msg = 'history repo dirty'
                else:
                    msg = 'history repo clean'
                self.log.info(msg)
                return Empty()
            self._with_current_repo(_ / log_status)

        @may_handle(HistoryLog)
        def history_log(self):
            self._with_current_repo(
                _ % (lambda r: self.vim.multi_line_info(r.log_formatted))
            )

        @handle(HistoryBrowse)
        def history_browse(self):
            def f(repo):
                commits = repo.history_info
                return ScratchBuilder().build.unsafe_perform_io(self.vim)\
                    .leffect(self._io_error)\
                    .map(lambda a: BrowseState(repo=repo, current=0,
                                               commits=commits, buffer=a))
            return self._current_repo_ro\
                .flat_map(f)\
                .map(self._add_browse)

        def _io_error(self, exc):
            self.log.caught_exception('nvim io', exc)

        @may_handle(Save)
        def save(self):
            return Commit()

        @property
        def _timestamp(self):
            return datetime.now().isoformat()

        @handle(HistoryBrowseInput)
        def history_browse_input(self):
            return self._current_browse.map(__.send(self.msg))

        @property
        def _current_browse(self):
            return self._browse_for_buffer(self.vim.buffer)

        def _browse_for_buffer(self, buffer):
            return self.state.browse\
                .find(_.buffer.buffer == buffer)\
                .map(_[1])

        @handle(QuitBrowse)
        def quit_browse(self):
            return self._browse_for_buffer(self.msg.buffer)\
                .map(self._remove_browse)

        def _add_browse(self, state: BrowseState):
            browse = Browse(state, self.vim)
            return (
                self._with_browse(self.state.browse + (browse.repo, browse)),
                IO(browse.run)
            )

        def _remove_browse(self, target: Browse):
            return self._with_browse(self.state.browse - target.repo)

        @handle(HistoryPick)
        def pick(self):
            return self._pick_commit\
                .flat_map(lambda a: self._pick_patch(a[0], a[1]))

        @handle(HistoryRevert)
        def revert(self):
            return self._pick_commit\
                .map(lambda a: self._pick_revert(a[0], a[1]))

        @property
        def _pick_commit(self):
            index = try_convert_int(self.msg.index)
            lifter = self._current_repo_ro / _.history_info / _.lift
            return index.ap(lifter).flatten.product(self.current)

        def _pick_patch(self, commit: CommitInfo, project: Project):
            patch = (
                commit.diff /
                _.revert //
                _.patch
            )
            apply = lambda pat: self._apply_patch(project, pat, commit)
            return patch.map(apply)

        async def _apply_patch(self, project, patch, commit):
            executor = Patch(self.vim)
            result = await executor.patch(project, patch)
            return self._check_pick_status(commit, executor, result)

        async def _pick_revert(self, commit: CommitInfo, project: Project):
            executor = HistoryGit(self.machine.base, self.vim)
            result = await executor.revert(project, commit)
            ret = self._check_pick_status(commit, executor, result)
            if not result.success:
                await executor.revert_abort(project)
            return ret

        @may
        def _check_pick_status(self, commit, executor, result):
            if result.success:
                self.vim.reload_windows()
                self.log.info('picked {}'.format(commit.num_since))
            else:
                return Error(Plugin.failed_pick_err).pub

__all__ = ('Commit', 'Plugin', 'HistoryPrev', 'HistoryNext',
           'HistoryBufferPrev', 'HistoryBufferNext', 'HistorySwitch',
           'HistoryPick', 'HistoryRevert')



import re
from pathlib import Path  # type: ignore

from fn import F, _  # type: ignore

from tryp import List, Map, Empty
from tryp.lazy import lazy

from trypnv.machine import handle, may_handle, message, Error  # type: ignore
from trypnv.process import JobClient

from proteome.state import ProteomeComponent
from proteome.project import Project, mkpath
from proteome.env import Env
from proteome.git import Git

StageI = message('StageI')
StageII = message('StageII')
StageIII = message('StageIII')
StageIV = message('StageIV')
Add = message('Add', 'project')
RemoveByIdent = message('RemoveByIdent', 'ident')
Create = message('Create', 'name', 'root')
Next = message('Next')
Prev = message('Prev')
SetProject = message('SetProject', 'ident')
SetProjectIdent = message('SetProjectIdent', 'ident')
SetProjectIndex = message('SetProjectIndex', 'index')
SwitchRoot = message('SwitchRoot', opt_fields=(('notify', True),))
Save = message('Save')
Added = message('Added', 'project')
Removed = message('Removed', 'project')
ProjectChanged = message('ProjectChanged', 'project')
BufEnter = message('BufEnter', 'buffer')
Initialized = message('Initialized')
MainAdded = message('MainAdded')
Show = message('Show', varargs='names')
AddByParams = message('AddByParams', 'ident', 'params')
CloneRepo = message('CloneRepo', 'uri', 'params')


class Plugin(ProteomeComponent):

    def _no_such_ident(self, ident: str, params):
        self.log.error('no project found matching \'{}\''.format(ident))

    @may_handle(StageI)
    def stage_1(self, env: Env, msg):
        main = env.analyzer(self.vim).main
        return Add(main), MainAdded().pub  # type: ignore

    @may_handle(StageIV)
    def stage_4(self, env, msg):
        return BufEnter(self.vim.buffer).pub, Initialized().pub

    @may_handle(Initialized)
    def initialized(self, env, msg):
        return env.set(initialized=True), SwitchRoot(False)

    @may_handle(MainAdded)
    def main_added(self, env, msg):
        env.main.foreach(self._setup_main)

    def _setup_main(self, pro: Project):
        self.vim.set_pvar('main_name', pro.name)
        self.vim.set_pvar('main_ident', pro.ident)
        self.vim.set_pvar('main_type', pro.tpe | 'none')
        self.vim.set_pvar('main_types', pro.all_types)

    @handle(AddByParams)
    def add_by_params(self, env: Env, msg):
        params = Map(msg.params)
        return params.get('root').map(mkpath)\
            .map(lambda a: env.loader.from_params(msg.ident, a, params))\
            .get_or_else(
                env.loader.by_ident(msg.ident)
                .or_else(env.loader.resolve_ident(
                    msg.ident, params, env.main_type)))\
            .map(Add)\
            .error(lambda: self._no_such_ident(msg.ident, params))

    @may_handle(Add)
    def add(self, env: Env, msg):
        if msg.project not in env:
            return env.add(msg.project), Added(msg.project).pub

    @may_handle(Added)
    def added(self, env, msg):
        self.vim.set_pvar('added_project', msg.project.json)
        self.vim.set_pvar('projects', env.projects.json)
        self.vim.pautocmd('Added')
        if env.initialized:
            return SetProjectIndex(-1)

    # TODO switch project if removed current
    @handle(RemoveByIdent)
    def remove_by_ident(self, env: Env, msg):
        return env.project(msg.ident)\
            .map(lambda a: (env - a, Removed(a).pub))

    @may_handle(Removed)
    def removed(self, env: Env, msg):
        self.vim.set_pvar('projects', env.projects.json)
        self.vim.pautocmd('Removed')

    @may_handle(Create)
    def create(self, env: Env, msg):
        return env + Project.of(msg.name, Path(msg.root))

    @may_handle(Show)
    def show(self, env: Env, msg):
        lines = env.projects.show(List.wrap(msg.names))
        header = List('Projects:')  # type: List[str]
        self.log.info('\n'.join(header + lines))

    @may_handle(SetProject)
    def set_project(self, env: Env, msg):
        if isinstance(msg.ident, str):
            if msg.ident in env:
                return SetProjectIdent(msg.ident)
            elif msg.ident.isdigit():
                return SetProjectIndex(int(msg.ident))
            else:
                err = '\'{}\' is not a valid project identifier'
                return Error(err.format(msg.ident))

    @may_handle(SetProjectIndex)
    def set_project_index(self, env, msg):
        if msg.index < env.project_count:
            return env.set_index(msg.index), SwitchRoot()

    @handle(SetProjectIdent)
    def set_project_ident(self, env: Env, msg):
        return env.set_index_by_ident(msg.ident)\
            .map(lambda a: (a, SwitchRoot()))

    @handle(SwitchRoot)
    def switch_root(self, env: Env, msg):
        def go(pro: Project):
            self.vim.switch_root(pro.root)  # type: ignore
            if msg.notify:
                info = 'switched root to {}'
                self.log.info(info.format(pro.ident))
            return ProjectChanged(pro)
        if env.initialized:
            return env.current.map(go)
        else:
            return Empty()

    @may_handle(ProjectChanged)
    def project_changed(self, env, msg):
        self.vim.set_pvar('active', msg.project.json)

    @may_handle(Next)
    def next(self, env: Env, msg):
        return env.inc(1), SwitchRoot()

    @may_handle(Prev)
    def prev(self, env: Env, msg):
        return env.inc(-1), SwitchRoot()

    @may_handle(Error)
    def error(self, env, msg):
        self.log.error(msg.message)

    # TODO make configurable (destination dir)
    @handle(CloneRepo)
    def clone_repo(self, env, msg):
        uri = msg.uri
        name = msg.params.get('name')\
            .or_else(self._clone_repo_name(uri))
        url = self._clone_url(uri)
        return (
            env.main_clone_dir.map2(name, _ / _)
            .to_either('invalid parameter: {}'.format(uri)) /
            F(self._clone_repo, url)
        )

    def _clone_url(self, uri):
        return uri if uri.startswith('http') else self._github_url(uri)

    def _github_url(self, uri):
        return 'https://github.com/{}'.format(uri)

    def _clone_repo_name(self, uri):
        return (
            List.wrap(uri.split('/'))
            .lift(-1) /
            F(re.sub, '\.git$', '')
        )

    @property
    def cloner(self) -> Git:
        return self._cloner  # type: ignore

    @lazy
    def _cloner(self):
        return Git(self.vim)

    async def _clone_repo(self, url: str, target):
        ident = '/'.join(str(target).split('/')[-2:])
        client = JobClient(cwd=Path.home(), name=self.name)
        res = await self.cloner.clone(client, url, target)
        return res.either(AddByParams(ident, {}).pub,
                          'failed to clone {} to {}'.format(url, target))

__all__ = ('Create', 'AddByParams', 'Plugin', 'Show', 'StageI', 'StageII',
           'StageIII', 'RemoveByIdent', 'Next', 'Prev', 'SetProjectIndex',
           'Save', 'Added', 'Removed', 'ProjectChanged', 'BufEnter',
           'Initialized', 'MainAdded', 'StageIV')



from pathlib import Path  # type: ignore

from toolz.itertoolz import cons  # type: ignore

from tryp import List, Map, Maybe, Empty

from proteome.nvim import NvimFacade
from proteome.env import Env
from proteome.state import ProteomeState
from proteome.project import Projects


class Proteome(ProteomeState):

    def __init__(
            self,
            vim: NvimFacade,
            config_path: Path,
            plugins: List[str],
            bases: List[Path],
            type_bases: Map[Path, List[str]],
            initial_projects: Maybe[Projects]=Empty()
    ) -> None:
        self._config_path = config_path
        self._bases = bases
        self._type_bases = type_bases
        self._initial_projects = initial_projects
        core = 'proteome.plugins.core'
        ProteomeState.__init__(self, vim, List.wrap(cons(core, plugins)))

    def init(self):
        return Env(  # type: ignore
            config_path=self._config_path,
            bases=self._bases,
            type_bases=self._type_bases,
            projects=self._initial_projects | Projects()
        )

    @property
    def projects(self):
        return self.env.projects

__all__ = ('Proteome',)



from trypnv.machine import message  # type: ignore

Commit = message('Commit')
HistorySwitch = message('HistorySwitch', 'index')
HistoryPrev = message('HistoryPrev')
HistoryNext = message('HistoryNext')
HistoryBufferPrev = message('HistoryBufferPrev')
HistoryBufferNext = message('HistoryBufferNext')
HistoryStatus = message('HistoryStatus')
HistoryLog = message('HistoryLog')
HistoryBrowse = message('HistoryBrowse')
HistoryBrowseInput = message('HistoryBrowseInput', 'keyseq')
HistoryPick = message('HistoryPick', 'index')
HistoryRevert = message('HistoryRevert', 'index')
Redraw = message('Redraw')
QuitBrowse = message('QuitBrowse', 'buffer')
ExecPick = message('ExecPick', 'commit', 'executor', 'status')
RevertAbort = message('ExecPick', 'project', 'executor', 'status')



from fn import F, _  # type: ignore

from trypnv.machine import may_handle

from proteome.state import ProteomeComponent
from proteome.env import Env
from proteome.plugins.core import MainAdded, StageIII


class Plugin(ProteomeComponent):

    @property
    def _project_dir(self):
        return self.vim.pvar('config_project_dir')\
            .get_or_else('project')

    @property
    def _project_after_dir(self):
        return self.vim.pvar('config_project_after_dir')\
            .get_or_else('project_after')

    def _runtime(self, project, base):
        run = F(self.vim.runtime) << F('{}/{}'.format, base)
        project.all_types.foreach(run)
        project.tpe.map(_ + '/' + project.name).foreach(run)
        run('all/*')

    def _runtime_before(self, project):
        return self._runtime(project, self._project_dir)

    def _runtime_after(self, project):
        return self._runtime(project, self._project_after_dir)

    @may_handle(MainAdded)
    def before(self, env: Env, msg):
        env.current.foreach(self._runtime_before)

    @may_handle(StageIII)
    def stage_3(self, env: Env, msg):
        env.current.foreach(self._runtime_after)

__all__ = ('Plugin',)



from pathlib import Path
import tempfile

from fn import _  # type: ignore

from proteome.project import (Projects, Resolver, ProjectLoader, Project,
                              ProjectAnalyzer)
from proteome.logging import Logging
from trypnv.data import Data  # type: ignore
from trypnv.record import field  # type: ignore
from trypnv import NvimFacade

from tryp import List, Map, Just


class Env(Data, Logging):
    config_path = field(Path)
    bases = field(List)
    type_bases = field(Map)
    projects = field(Projects)
    current_index = field(int, initial=0)
    initialized = field(bool, initial=False)

    @property
    def loader(self):
        return ProjectLoader(self.config_path, self.resolver)

    @property
    def resolver(self):
        return Resolver(self.bases, self.type_bases)

    def __str__(self):
        return '{}({},{},{})'.format(
            self.__class__.__name__,
            self.current_index,
            self.initialized,
            self.projects,
        )

    @property
    def current(self):
        return self.projects[self.current_index]

    @property
    def project_count(self):
        return len(self.projects)

    def inc(self, num):
        new_index = (self.current_index + num) % self.project_count
        return self.set_index(new_index)

    def set_index(self, index):
        return self.set(current_index=index)

    def set_index_by_ident(self, ident):
        return self.projects.index_of_ident(ident)\
            .map(self.set_index)

    def add(self, pro: Project):
        return self if pro in self else self.set(projects=self.projects + pro)

    def __add__(self, pro: Project):
        return self.add(pro)

    def remove(self, ident):
        pro = (Just(ident) if isinstance(ident, Project) else
               self.project(ident))
        return pro\
            .cata(lambda p: self.set(projects=self.projects - p), self)

    def __sub__(self, ident):
        return self.remove(ident)

    def analyzer(self, vim: NvimFacade):
        return ProjectAnalyzer(vim, self.loader)

    def project(self, ident: str):
        return self.projects.project(ident)

    @property
    def all_projects(self):
        return self.projects.projects

    def __contains__(self, item):
        return item in self.projects

    @property
    def main(self):
        return self.projects[0]

    def history_projects(self, all_projects: bool):
        want_history = _.history
        filter = ((lambda a: want_history(a) or a.has_type)
                  if all_projects
                  else want_history)
        return self.all_projects.filter(filter)

    @property
    def main_type(self):
        return self.main.flat_map(_.tpe)

    @property
    def addable(self):
        return self.loader.all_ident(self.main_type)

    @property
    def main_addable(self):
        return self.loader.main_ident(self.main_type)

    @property
    def main_clone_dir(self):
        temp = lambda: tempfile.mkdtemp(prefix='proteome_clone')
        return self.resolver.bases.head\
            .map2(self.main_type, _ / _)\
            .or_else(temp)

__all__ = ['Env']



from fn import _  # type: ignore

from trypnv.machine import may_handle, message, handle  # type: ignore
from trypnv.nvim import Buffer

from tryp import List, Empty
from tryp.lazy import lazy

from proteome.state import ProteomeComponent
from proteome.env import Env
from proteome.ctags import Ctags
from proteome.plugins.core import Save, Added, BufEnter, StageIV

Gen = message('Gen', 'project')
GenAll = message('GenAll')
Kill = message('Kill')
CurrentBuffer = message('CurrentBuffer')


class Plugin(ProteomeComponent):

    @property
    def ctags(self) -> Ctags:
        return self._ctags  # type: ignore

    @lazy
    def _ctags(self):
        return Ctags(self.vim)

    @property
    def _tags_file_name(self):
        return '.tags'

    @handle(Gen)
    async def gen(self, env, msg):
        await self.ctags.gen(msg.project)
        return Empty()

    @may_handle(StageIV)
    def stage_4(self, env: Env, msg):
        return GenAll(), CurrentBuffer()

    @may_handle(Save)
    def save(self, env: Env, msg):
        return GenAll()

    @may_handle(GenAll)
    def gen_all(self, env: Env, msg):
        if self.ctags.ready:
            return env.projects.projects\
                .filter(_.want_ctags)\
                .map(Gen)

    # TODO kill dangling procs
    @may_handle(Kill)
    def kill(self, env: Env, msg):
        pass

    @may_handle(BufEnter)
    def buf_enter(self, env, msg):
        self.set_buffer_tags(env, List(msg.buffer))

    @may_handle(Added)
    def added(self, env: Env, msg: Added):
        return CurrentBuffer()

    @may_handle(CurrentBuffer)
    def buffer(self, env: Env, msg):
        self.set_buffer_tags(env, List(self.vim.buffer))

    def set_buffer_tags(self, env: Env, bufs: List[Buffer]):
        files = env.all_projects.map(_.root / self._tags_file_name)
        bufs.foreach(lambda a: a.amend_optionl('tags', files))

__all__ = ('GenAll', 'Plugin')



from setuptools import setup, find_packages  # type: ignore

setup(
    name='proteome',
    description='project management for neovim',
    version='6.7.0',
    author='Torsten Schmits',
    author_email='torstenschmits@gmail.com',
    license='MIT',
    url='https://github.com/tek/proteome',
    packages=find_packages(
        exclude=['unit', 'unit.*', 'integration', 'integration.*']),
    install_requires=[
        'tryp-nvim>=6.3.0',
        'tryp>=6.5.0',
        'dulwich',
    ]
)
